@@ -14,6 +14,7 @@ from isolate.backends import (
 from isolate.connections._local import PythonExecutionBase, agent_startup
 from isolate.connections.common import serialize_object
 from isolate.connections.grpc import agent, definitions
+from isolate.connections.grpc.configuration import get_default_options
 from isolate.connections.grpc.interface import from_grpc, to_grpc
 from isolate.logs import Log, LogLevel, LogSource
 
@@ -31,20 +32,36 @@ class GRPCExecutionBase(EnvironmentConnection):
         the required credentials to connect to it."""
         raise NotImplementedError
 
-    def _run_through_grpc(
+    @contextmanager
+    def _establish_bridge(
         self,
-        function: definitions.SerializedObject,
         *,
-        max_wait_timeout: float = 10,
-    ) -> Iterator[definitions.PartialRunResult]:
-        """Starts a new agent process, sends the given 'function' to the agent and
-        awaits until the result is received. This function takes and returns raw gRPC
-        messages (use `run()` to actually execute Python functions and receive Python
-        objects).
-
-        If the gRPC server is not accessible at the passed maximum timeout, an
-        AgentError will be raised."""
+        max_wait_timeout: float = 10.0,
+    ) -> Iterator[definitions.AgentStub]:
+        with self.start_agent() as (address, credentials):
+            with grpc.secure_channel(
+                address,
+                credentials,
+                options=get_default_options(),
+            ) as channel:
+                channel_status = grpc.channel_ready_future(channel)
+                try:
+                    channel_status.result(timeout=max_wait_timeout)
+                except grpc.FutureTimeoutError:
+                    raise AgentError(
+                        f"Couldn't connect to the gRPC server in the agent (listening at {address}) "
+                        "in time."
+                    )
+                stub = definitions.AgentStub(channel)
+                stub._channel = channel  # type: ignore
+                yield stub
 
+    def run(
+        self,
+        executable: BasicCallable,
+        *args: Any,
+        **kwargs: Any,
+    ) -> CallResultType:
         # Implementation details
         # ======================
         #
@@ -63,26 +80,6 @@ class GRPCExecutionBase(EnvironmentConnection):
         #                   using the same serialization method.
         #  8. [controller]: Receive the result back and return it.
 
-        with self.start_agent() as (address, credentials):
-            with grpc.secure_channel(address, credentials) as channel:
-                channel_status = grpc.channel_ready_future(channel)
-                try:
-                    channel_status.result(timeout=max_wait_timeout)
-                except grpc.FutureTimeoutError:
-                    raise AgentError(
-                        f"Couldn't connect to the gRPC server in the agent (listening at {address}) "
-                        "in time."
-                    )
-
-                agent = definitions.AgentStub(channel)
-                yield from agent.Run(function)
-
-    def run(
-        self,
-        executable: BasicCallable,
-        *args: Any,
-        **kwargs: Any,
-    ) -> CallResultType:
         method = self.environment.settings.serialization_method
         function = definitions.SerializedObject(
             method=method,
@@ -90,19 +87,23 @@ class GRPCExecutionBase(EnvironmentConnection):
             was_it_raised=False,
             stringized_traceback=None,
         )
+        function_call = definitions.FunctionCall(
+            function=function,
+        )
 
-        for partial_result in self._run_through_grpc(function):
-            for raw_log in partial_result.logs:
-                log = from_grpc(raw_log)
-                self.log(log.message, level=log.level, source=log.source)
+        with self._establish_bridge() as bridge:
+            for partial_result in bridge.Run(function_call):
+                for raw_log in partial_result.logs:
+                    log = from_grpc(raw_log)
+                    self.log(log.message, level=log.level, source=log.source)
 
-            if partial_result.is_complete:
-                if not partial_result.result:
-                    raise AgentError(
-                        "The agent didn't return a result, but it should have."
-                    )
+                if partial_result.is_complete:
+                    if not partial_result.result:
+                        raise AgentError(
+                            "The agent didn't return a result, but it should have."
+                        )
 
-                return cast(CallResultType, from_grpc(partial_result.result))
+                    return cast(CallResultType, from_grpc(partial_result.result))
 
         raise AgentError(
             "No result object was received from the agent (it never set is_complete to True)."
